#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "iscm0.h"

static OBJ* funcall(OBJ* op, OBJ* params, void* env);

/**** environment ****/

typedef struct S_KVPAIR {
  const char* key;
  OBJ* val;
  struct S_KVPAIR* next;
} KEY_VALUE_PAIR;

typedef struct S_SCM_ENV {
  KEY_VALUE_PAIR* kvs;
  struct S_SCM_ENV* parent;
} SCM_ENV;

static SCM_ENV root_env = {NULL, NULL};

static int
lookup(SCM_ENV* env, const char *const key, OBJ** pp)
{
  KEY_VALUE_PAIR *kvs;

  if (!env){ env = &root_env; }

  kvs = env->kvs;
  while (kvs){
    if (kvs->key && key && strcmp(kvs->key, key) == 0){
      *pp = kvs->val;
      return 1;
    }
    kvs = kvs->next;
  }

  return (env->parent && lookup(env->parent, key, pp));
}

static void
define_var(SCM_ENV* env, const char *const key, OBJ* exp)
{
  KEY_VALUE_PAIR *kv;

  if (!env){ env = &root_env; }

  kv = mycalloc(sizeof(KEY_VALUE_PAIR));
  kv->key = key;
  kv->val = exp;
  kv->next = env->kvs;
  env->kvs = kv;
}

static void
set_var(SCM_ENV* env, const char *const key, OBJ* exp)
{
  KEY_VALUE_PAIR *kvs;

  if (!env){ env = &root_env; }

  kvs = env->kvs;
  while (kvs){
    if (kvs->key && key && strcmp(kvs->key, key) == 0){
      kvs->val = exp;
      return;
    }
    kvs = kvs->next;
  }

  if (env->parent){
    set_var(env->parent, key, exp);
  } else {
    fprintf(stderr, "Error: Unbound variable: %s\n",key);
    exit(1);
  }
}

/**** built-in functions ****/

static void
pair_chk(const OBJ* exp, const char *const opname)
{
  if (!exp || exp->ty != PAIR_OBJ){
    print_obj(exp);
    fprintf(stderr, "Fatal: %s needs a pair.\n", opname);
    exit(1);
  }
}

static OBJ* car(const OBJ* exp) { pair_chk(exp, "car"); return exp->u.car; }
static OBJ* cdr(const OBJ* exp) { pair_chk(exp, "cdr"); return exp->cdr; }

static int
id_comp(OBJ* exp, const char *const str)
{
  return (exp->ty == ID_OBJ && strcmp(exp->u.pstr, str) == 0);
}

enum {CAR, CDR, DEFINE, LAMBDA, SET, COND, QUOTE,
      ADD, SUB, MUL, DIV, EQ, GT, DISP, VER, not_found};

static int
check_builtin_op(OBJ* op)
{
  int ret = not_found;

  if      (id_comp(op, "car"))     { ret = CAR; }
  else if (id_comp(op, "cdr"))     { ret = CDR; }
  else if (id_comp(op, "define"))  { ret = DEFINE; }
  else if (id_comp(op, "lambda"))  { ret = LAMBDA; }
  else if (id_comp(op, "set!"))    { ret = SET; }
  else if (id_comp(op, "cond"))    { ret = COND; }
  else if (id_comp(op, "quote"))   { ret = QUOTE; }
  else if (id_comp(op, "+"))       { ret = ADD; }
  else if (id_comp(op, "-"))       { ret = SUB; }
  else if (id_comp(op, "*"))       { ret = MUL; }
  else if (id_comp(op, "/"))       { ret = DIV; }
  else if (id_comp(op, "="))       { ret = EQ; }
  else if (id_comp(op, ">"))       { ret = GT; }
  else if (id_comp(op, "display")) { ret = DISP; }
  else if (id_comp(op, "iscm-version")) { ret = VER; }

  return ret;
}

OBJ*
disp(OBJ* params, void* env)
{
  OBJ* x;
  x = eval(car(params), env);
  if (!x){ printf("#<nil>"); }
  else if (x->ty == INT_OBJ){ printf("%ld", x->u.ival); }
  else if (x->ty == STR_OBJ){ printf(x->u.pstr); }
  else { print_obj(x); }
  return NULL;
}

static void
int_chk(OBJ* exp)
{
  if (!exp || exp->ty != INT_OBJ){
    print_obj(exp);
    fprintf(stderr, "Arithmetic operator needs integer(s).\n");
    exit(1);
  }
}

OBJ*
four_arithmetic(int op, OBJ* params, void* env)
{
  OBJ *ret, *x, *rest;

  ret = mycalloc(sizeof(OBJ));
  ret->ty = INT_OBJ;

  int_chk(x = eval(car(params), env));
  ret->u.ival = x->u.ival;

  rest = cdr(params);

  while (rest){
    int_chk(x = eval(car(rest), env));

    switch(op){
    case ADD:
      ret->u.ival += x->u.ival; break;
    case SUB:
      ret->u.ival -= x->u.ival; break;
    case MUL:
      ret->u.ival *= x->u.ival; break;
    case DIV:
      if (x->u.ival == 0){
	fprintf(stderr, "Divided by zero.\n");
	exit(1);
      }
      ret->u.ival /= x->u.ival;
    }

    rest = cdr(rest);
  }

  return ret;
}

static OBJ*
def(OBJ* params, void* env)
{
  OBJ *id, *exp;
  id = car(params);
  
  exp = eval(car(cdr(params)), env);

  if (id && id->ty == ID_OBJ){
    define_var(env, id->u.pstr, exp);
  } else {
    fprintf(stderr, "Syntax error on `define.'\n");
    exit(1);
  }

  return id;
}

static OBJ*
set(OBJ* params, void* env)
{
  OBJ *id, *exp;
  id = car(params);
  exp = eval(car(cdr(params)), env);
  set_var(env, id->u.pstr, exp);
  return exp;
}

static OBJ*
eq(OBJ* params, void* env)
{
  OBJ* ret, *x, *y, *rest;
  
  ret = mycalloc(sizeof(OBJ));
  ret->ty = INT_OBJ;
  ret->u.ival = 1;

  int_chk(x = eval(car(params), env));

  rest = cdr(params);
  while (rest){
    int_chk(y = eval(car(rest), env));
    if (x->u.ival != y->u.ival) { return NULL; }
    rest = cdr(rest);
  }

  return ret;
}

static OBJ*
gt(OBJ* params, void* env)
{
  OBJ* ret, *x, *y;

  ret = mycalloc(sizeof(OBJ));
  ret->ty = INT_OBJ;
  ret->u.ival = 1;

  int_chk(x = eval(car(params), env));
  int_chk(y = eval(car(cdr(params)), env));

  if (x->u.ival > y->u.ival){
    return ret;
  }

  return NULL;
}

static OBJ*
cond(OBJ* params, void* env)
{
  OBJ *ret = NULL;
  OBJ *clause, *tst, *exps, *exp;

  while(params){
    clause = car(params);
    params = cdr(params);

    tst = car(clause);
    exps = cdr(clause);

    if((tst->ty == ID_OBJ && strcmp(tst->u.pstr, "else") == 0) ||
       eval(tst, env)){
      while(exps && (exp = car(exps))){
	ret = eval(exp, env);
	exps = cdr(exps);
      }
      break;
    }
  }

  return ret;
}

static OBJ builtin_func = { BUILTIN_FUNC, {0}, NULL };

/**** closure ****/
typedef struct {
  SCM_ENV* env;
  const OBJ* dummy_args;
  OBJ* exp;
} CLOSURE;

static OBJ*
lambda(OBJ* params, void* parent_env)
{
  OBJ* obj;
  CLOSURE *closure;

  obj = mycalloc(sizeof(OBJ));
  closure = mycalloc(sizeof(CLOSURE));

  if(parent_env){
    closure->env = parent_env;
  } else {
    closure->env = &root_env;
  }
  closure->dummy_args = car(params);
  closure->exp = cdr(params);
  
  obj->ty = FUNC;
  obj->u.pclosure = closure;

  return obj;
}

/**** eval ****/

static OBJ version_str_obj = {STR_OBJ, {ISCM_VERSION}, NULL};

OBJ*
eval(OBJ* exp, void* env)
{
  OBJ* ret = NULL;
  OBJ *obj, *op;
  /* print_obj(exp); */

  if (!exp || exp->ty == INT_OBJ || exp->ty == STR_OBJ || exp->ty == FUNC){
    return exp;
  }

  if (exp && exp->ty == ID_OBJ){
    if (lookup(env, exp->u.pstr, &obj)){
      return eval(obj, env);
    } else if (check_builtin_op(exp) != not_found) {
      return &builtin_func;
    } else {
      fprintf(stderr, "Unbound variable: (%s)\n", exp->u.pstr);
      exit(1);
    }
  }

  op = eval(car(exp), env);

  if (op && op->ty == BUILTIN_FUNC){
    switch(check_builtin_op(car(exp))){
    case CAR:    ret = car(eval(car(cdr(exp)), env));       break;
    case CDR:    ret = cdr(eval(car(cdr(exp)), env));       break;
    case DEFINE: ret = def(cdr(exp), env);                  break;
    case SET:    ret = set(cdr(exp), env);                  break;
    case COND:   ret = cond(cdr(exp), env);                 break;
    case LAMBDA: ret = lambda(cdr(exp), env);               break;
    case QUOTE:  ret = cdr(exp);                            break;
    case ADD:    ret = four_arithmetic(ADD, cdr(exp), env); break;
    case SUB:    ret = four_arithmetic(SUB, cdr(exp), env); break;
    case MUL:    ret = four_arithmetic(MUL, cdr(exp), env); break;
    case DIV:    ret = four_arithmetic(DIV, cdr(exp), env); break;
    case EQ:     ret = eq(cdr(exp), env);                   break;
    case GT:     ret = gt(cdr(exp), env);                   break;
    case DISP:   ret = disp(cdr(exp), env);                 break;
    case VER:    ret = &version_str_obj;                    break;
    default:     fprintf(stderr, "Operator not found.\n");  break;
    }
  } else if (op && op->ty == FUNC){
    ret = funcall(op, cdr(exp), env);
  } else {
    print_obj(exp);
    fprintf(stderr, "Error: not an operator.\n");
    exit(1);
  }

  return ret;
}

/**** function call ****/

static SCM_ENV*
setup_parameter_env(const OBJ *dummy_args, OBJ *params, void* env)
{
  OBJ *k, *v;
  SCM_ENV *ret;
  ret = mycalloc(sizeof(SCM_ENV));

  while (dummy_args && dummy_args->ty == PAIR_OBJ &&
	 params && params->ty == PAIR_OBJ){
    k = car(dummy_args);
    v = car(params);
    define_var(ret, k->u.pstr, eval(v, env));
    dummy_args = cdr(dummy_args);
    params = cdr(params);
  }

  return ret;
}

static OBJ*
funcall(OBJ* op, OBJ* params, void* env)
{
  OBJ *ret = NULL, *exp;
  CLOSURE* closure;
  SCM_ENV* local_env;
  SCM_ENV* org_parent;

  closure = op->u.pclosure;
  
  local_env = setup_parameter_env(closure->dummy_args, params, env);
  
  org_parent = closure->env;
  closure->env = local_env;
  local_env->parent = org_parent;

  exp = closure->exp;
  while (exp){
    ret = eval(car(exp), closure->env);
    exp = cdr(exp);
  }

  return ret;
}
